-- Copyright (C) 2011 Jared Krinke.
-- 
-- This is free software; you can redistribute it and/or
-- modify it under the terms of the GNU General Public License
-- as published by the Free Software Foundation; either version 2
-- of the License, or (at your option) any later version.
-- 
-- This program is distributed in the hope that it will be useful,
-- but WITHOUT ANY WARRANTY; without even the implied warranty of
-- MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
-- GNU General Public License for more details.
-- 
-- You should have received a copy of the GNU General Public License
-- along with this program; if not, write to the Free Software
-- Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.

-- The Grind
require("UI/Input.lua")
require("UI/Layers.lua")

local debug = false
local edit = false

-- Constants
Grind = {
    version = "0.1",
    textWidth = 6,
    textHeight = 8,
}

Grind.updateOrder = {
    master = 0,
}

--Form.setDefaultTextSize(Grind.textWidth, Grind.textHeight)

-- Controls
local logicalEvents = {
    up = { necessary = true },
    down = { necessary = true },
    left = { necessary = true },
    right = { necessary = true },
    activate = { necessary = true },
    cancel = { necessary = true },
    switch = { necessary = true },
    shift = { necessary = false },
}

-- TODO: Move to settings file
local controlsData = {
    up = "up",
    down = "down",
    left = "left",
    right = "right",
    space = "activate",
    enter = "activate",
    escape = "cancel",
    tab = "switch",
    lshift = "shift",
    rshift = "shift",
}

Grind.input = Input.new(logicalEvents, controlsData)
--FormLayer.setDefaultInput(Grind.input)

-- Set default video mode
-- TODO: Move to settings file
Video.setMode(800, 600, false)
--Audio.setVolume(1)

TaskList = { }

function TaskList.new()
    local taskList = { }

    taskList.tasks = { }
    taskList.p = { }

    function taskList.add(self, p, init, run, cleanup)
        local index = self:addInternal(init, run, cleanup)
        local i = 1

        while i <= p do
            self.p[#self.p + 1] = index
            i = i + 1
        end
    end

    function taskList.addInternal(self, init, run, cleanup)
        local entry = { init, run, cleanup }

        self.tasks[#self.tasks + 1] = entry

        return #self.tasks
    end

    function taskList.force(self, o, index)
        -- Cleanup last task
        if o.taskListIndex ~= nil then
            local entry = self.tasks[o.taskListIndex]

            if isFunction(entry[3]) then
                entry[3](o)
            end
        end

        -- Setup the new task
        local entry = self.tasks[index]

        if isFunction(entry[1]) then
            entry[1](o)
        end

        o.taskListIndex = index

        return entry[2]
    end

    function taskList.choose(self, o)
        return self:force(o, random(#self.p) + 1)
    end

    return taskList
end

Actor = {
    width = 1,
    -- TODO: With 3/4 perspective, this should be 4
    height = 2,

    type = {
        player = 1,
        boss = 2,
    }
}

Actor.names = {
    [Actor.type.player] = "Player",
    [Actor.type.boss] = "Boss",
}

local function createAnimations(name)
    return {
        stand = Animation.new({ String.format("Images/%sStand.png", name), 0 }),
    }
end

Actor.animations = { }
Actor.animations[Actor.type.player] = createAnimations(Actor.names[Actor.type.player])
Actor.animations[Actor.type.boss] = createAnimations(Actor.names[Actor.type.boss])

function Actor.new(actorType, x, y)
    local actor = Entity.new(x, y)

    actor.actorType = actorType
    actor.dx, actor.dy = 0, 0

    -- Set up animations
    local animationElements = { }

    Table.forEach(Actor.animations[actorType], function(name, animation)
        -- TODO: Z
        animationElements[name] = Element.Animation.new(animation, 0, Actor.height / 2, 0, Actor.width, Actor.height)
    end)

    actor.animationElements = animationElements

    function actor.setAnimation(self, animationName)
        local animationElement = self.animationElements[animationName]

        -- TODO: Animation reset (only in Radius Engine 0.8+) isn't needed for v0.1
        --animationElement:reset()
        self.elements:clear()
        self.elements:add(animationElement)
    end

    function actor.move(self, speed, ms)
        local oldX, oldY = self.x, self.y

        -- Determine component speeds
        if self.dx ~= 0 or self.dy ~= 0 then
            local angle = arctan(self.dy, self.dx)
            local dx = cos(angle) * speed
            local dy = sin(angle) * speed

            -- Move
            local x, y = self.x + dx * ms, self.y + dy * ms

            self.x = self.master:getValidX(self, x)
            self.y = self.master:getValidY(self, y)
        end

        return self.x ~= oldX or self.y ~= oldY
    end

    actor:setAnimation("stand")

    return actor
end

local directions = {
    { 0, 1 },
    { 0, -1 },
    { -1, 0 },
    { 1, 0 },
}

local function angleToDirection(angle)
    if angle >= -45 and angle < 45 then
        return directions[4]
    elseif angle >= 45 and angle < 135 then
        return directions[1]
    elseif angle >= 135 or angle < -135 then
        return directions[3]
    else
        return directions[2]
    end
end

Boss = {
    speed = 3 / 1000,
    tasks = TaskList.new(),
    awareFront = 8,
    awareBack = 2,
    line = "Have some work!"
}

-- Stand (1 sec)
Boss.tasks:add(1, function(self)
        self.dx, self.dy = 0, 0
        self.taskTimer = 1000
    end, function(self, ms)
        self.taskTimer = self.taskTimer - ms
        return self.taskTimer > 0
    end)

-- Follow path (2 sec)
Boss.tasks:add(1, function(self)
        if not isNumber(self.landmark) then
            self.landmark = 0
        end

        self.taskTimer = 2000
    end, function(self, ms)
        local continue = false

        if #self.landmarks > 0 then
            -- Check for completion of this step
            local landmark = self.landmarks[self.landmark]

            if self.landmark == 0 or (self.x == landmark[1] and self.y == landmark[2]) then
                self.landmark = (self.landmark % #self.landmarks) + 1
                landmark = self.landmarks[self.landmark]
                self.landmarkX = landmark[1] - self.x
                self.landmarkY = landmark[2] - self.y
            end

            self.dx = landmark[1] - self.x
            self.dy = landmark[2] - self.y
            continue = self:move(Boss.speed, ms)

            -- Check for moving too far
            if self.landmarkX > 0 then
                self.x = min(self.x, landmark[1])
            elseif self.landmarkX < 0 then
                self.x = max(self.x, landmark[1])
            end

            if self.landmarkY > 0 then
                self.y = min(self.y, landmark[2])
            elseif self.landmarkY < 0 then
                self.y = max(self.y, landmark[2])
            end

            if self.x == landmark[1] and self.y == landmark[2] then
                continue = false
            end
        end

        self.taskTimer = self.taskTimer - ms

        return continue and self.taskTimer > 0
    end)

-- Give work (5 sec)
Boss.workTaskIndex = 3
Boss.tasks:addInternal(function(self)
        self.dx, self.dy = 0, 0
        self.taskTimer = 5000
        self:addChild(self.line)
        self.master:addWork()
        self.active = false
    end, function(self, ms)
        local lastTimer = self.taskTimer

        self.taskTimer = self.taskTimer - ms

        if lastTimer > 3500 and self.taskTimer <= 3500 then
            self:removeChild(self.line)
        end

        return self.taskTimer > 0
    end, function(self)
        self:removeChild(self.line)
        self.active = true
    end)

function Boss.new(master, x, y)
    local boss = Actor.new(Actor.type.boss, x, y)

    boss.x, boss.y = x, y
    boss.master = master
    boss.task = Boss.tasks:choose(boss)
    boss.lastAngle = 0
    boss.landmarks = { }
    boss.active = true

    local line = Entity.new()
    local textWidth = 0.33333
    local textHeight = 0.5

    line.elements:add(Element.Image.new("Images/Blank.png", 0, 1, 1, textWidth * #Boss.line, textHeight))
    line.elements:add(Element.Text.new(Boss.line, -textWidth * #Boss.line / 2, 1 - textHeight / 2, 2, textWidth, textHeight))
    boss.line = line

    if debug then
        local tmp = Entity.new()

        boss:addChild(tmp)
        boss.scanFront = tmp.elements:add(Element.Image.new("Images/Blank.png", 0, 0, 0, Boss.awareFront, Boss.awareFront, 0, Color.new(1, 0, 0, 0.25)))
        boss.scanBack = tmp.elements:add(Element.Image.new("Images/Blank.png", 0, 0, 0, Boss.awareBack, Boss.awareBack, 0, Color.new(1, 0, 0, 0.25)))
    end

    function boss.addLandmark(self, x, y)
        boss.landmarks[#boss.landmarks + 1] = { x, y }
    end

    function boss.update(self, ms)
        if not isFunction(boss.task) or not self:task(ms) then
            self.task = Boss.tasks:choose(self)
        end

        -- Record last direction
        if self.dx ~= 0 or self.dy ~= 0 then
            self.lastAngle = arctan(self.dy, self.dx)
        end
    end

    return boss
end

Player = {
    --stand = Element.Animation.new(Animation.new({ "Images/PlayerStand.png", 0})),
    --walkRight = Element.Animation.new(Animation.new(
    --    { "Images/PlayerWalkRight2.png", 150 },
    --    { "Images/PlayerStand.png", 150 },
    --    { "Images/PlayerWalkRight1.png", 150 },
    --    { "Images/PlayerStand.png", 150 },
    --    true)),
    speed = 6 / 1000,
}

function Player.new(master, x, y)
    local player = Actor.new(Actor.type.player, x, y)

    player.master = master
    player.z = 0.5
    player.pressed = { up = false, down = false, left = false, right = false }

    function player.setControlPressed(self, event, pressed)
        self.pressed[event] = pressed
    end

    function player.update(self, ms)
        -- Determine direction
        self.dx, self.dy = 0, 0

        if self.pressed.up then
            self.dy = 1
        elseif self.pressed.down then
            self.dy = -1
        end

        if self.pressed.right then
            self.dx = 1
        elseif self.pressed.left then
            self.dx = -1
        end

        self:move(Player.speed, ms)
    end

    return player
end

-- Default wall config
if edit then
    include("Config.lua")
else
    require("TheGrind/Map.lua")
end

Master = {
    boardWidth = 20,
    boardHeight = 20,
    minuteMS = 250,
    hoursOffset = 9,
}

local function createEmptyConfig()
    config = { }
    
    do
        local i = 1
    
        while i <= Master.boardHeight * 2 + 1 do
            local j = 1
    
            config[i] = { }
    
            while j <= Master.boardWidth * 2 + 1 do
                config[i][j] = 0
                j = j + 1
            end
    
            i = i + 1
        end
    end

    return config
end

local actorConstructors = {
    [Actor.type.player] = function(master, entry) return Player.new(master, entry[2], entry[3]) end,
    [Actor.type.boss] = function(master, entry)
            local boss = Boss.new(master, entry[2], entry[3])
            local landmarks = entry[4]
            local i = 1

            while i <= #landmarks do
                local landmark = landmarks[i]

                boss:addLandmark(landmark[1], landmark[2])
                i = i + 1
            end

            return boss
        end,
}

-- Set up empty config, if necessary
if config == nil or #config ~= Master.boardHeight * 2 + 1 or #config[1] ~= Master.boardWidth * 2 + 1 then
    config = createEmptyConfig()
    
    do
        local i = 1
        while i <= Master.boardHeight * 2 + 1 do
            local j = 1
            while j <= Master.boardWidth * 2 + 1 do
                if i == 1 or i == Master.boardHeight * 2 + 1 or j == 1 or j == Master.boardWidth * 2 + 1 then
                    config[i][j] = 1
                else
                    config[i][j] = 0
                end
    
                j = j + 1
            end
    
            i = i + 1
        end
    end
end

function Master.new(layer, x, y, width, height, completed)
    local master = Entity.new(x, y, 0, width / Master.boardWidth, height / Master.boardHeight)

    master.order = Grind.updateOrder.master
    master.completed = completed
    master.running = false

    -- Setup empty board
    local board = { }
    local wallsH = { }
    local wallsV = { }
    local i = 1

    local function createSpace(i, j)
        local space = Entity.new(i - 0.5, j - 0.5, 0, 1, 1)

        space.elements:add(Element.Image.new("Images/Blank.png"))
        space.color = Color.new(0.6, 0.6, 0.8)

        return space
    end

    local function createWall(i, j, horizontal)
        local wallWidth = 0.25
        local x = j - 1
        local y = i - 1

        local wall = Entity.new(x, y, -y / Master.boardHeight)

        -- Draw the wall
        -- TODO: Try and make walls look good (and succeed this time)
        --if debug then
            local width = wallWidth
            local height = 1
            local wx = 0
            local wy = 0

            if horizontal then
                width = 1
                height = wallWidth
                wx = 0.5
            else
                wy = 0.5
            end

            local e = wall.elements:add(Element.Image.new("Images/Blank.png", wx, wy, 0, width, height))
            e.color = Color.new(0.9, 0.6, 0.2)
        --else
        --    if horizontal then
        --        if i ~= 1 then
        --            wall.elements:add(Element.Image.new("Images/WallH.png", Actor.width / 2, Actor.height / 2, 0, Actor.width, Actor.height))
        --        end
        --    else
        --        wall.elements:add(Element.Image.new("Images/WallV.png", 0, Actor.height / 2, 0, wallWidth, Actor.height))
        --    end

        --    -- TODO: Does this look good?
        --    --local v = 1 - (i / Master.boardHeight) / 4

        --    --wall.color = Color.new(v, v, v)
        --end

        return wall
    end

    function master.addBoard(self)
        local i = 1

        while i <= Master.boardHeight + 1 do
            local j = 1

            wallsH[i] = { }
            wallsV[i] = { }

            if i <= Master.boardHeight then
                board[i] = { }
            end

            while j <= Master.boardWidth + 1 do
                if i <= Master.boardHeight and j <= Master.boardWidth then
                    board[i][j] = 0
                end

                -- Horizontal walls
                if j <= Master.boardWidth then
                    local value = config[i * 2 - 1][j * 2]

                    wallsH[i][j] = nil

                    if value == 1 then
                        wallsH[i][j] = self:addChild(createWall(i, j, true))
                    end
                end

                -- Vertical walls
                if i <= Master.boardHeight then
                    local value = config[i * 2][j * 2 - 1]

                    wallsV[i][j] = nil

                    if value == 1 then
                        wallsV[i][j] = self:addChild(createWall(i, j, false))
                    end
                end

                j = j + 1
            end

            i = i + 1
        end
    end

    function master.toggleWall(self, i, j, horizontal)
        local walls = wallsV

        if horizontal then
            walls = wallsH
        end

        local current = walls[i][j]

        -- Toggle
        if current == nil then
            walls[i][j] = self:addChild(createWall(i, j, horizontal))
        else
            self:removeChild(current)
            walls[i][j] = nil
        end
    end

    function master.save(self, filename)
        -- Create the config
        local config = createEmptyConfig()
        local i = 1
        while i <= Master.boardHeight + 1 do
            local j = 1
            while j <= Master.boardWidth + 1 do
                if j <= Master.boardWidth then
                    if wallsH[i][j] ~= nil then
                        config[i * 2 - 1][j * 2] = 1
                    end
                end

                if i <= Master.boardHeight then
                    if wallsV[i][j] ~= nil then
                        config[i * 2][j * 2 - 1] = 1
                    end
                end

                j = j + 1
            end
            i = i + 1
        end

        -- Save walls
        local file = File.openWrite(filename)

        file:write("config = {")
        i = 1

        while i <= #config do
            local buffer = StringBuffer.new()
            local j = 1

            buffer:append("    {")

            while j <= #config[i] do
                buffer:append(String.format(" %d,", config[i][j]))
                j = j + 1
            end

            buffer:append(" },")
            file:write(buffer.text)

            i = i + 1
        end

        file:write("}")

        -- Save actors
        file:write("actors = {")
        self:forEachChild(function(e)
            if isNumber(e.actorType) then
                if e.actorType == Actor.type.boss then
                    local buffer = StringBuffer.new()
                    local i = 1

                    buffer:append("{ ")

                    while i <= #e.landmarks do
                        local landmark = e.landmarks[i]

                        buffer:append(String.format("{ %f, %f }, ", landmark[1], landmark[2]))
                        i = i + 1
                    end

                    buffer:append("}")

                    file:write(String.format("    { %d, %f, %f, %s },", e.actorType, e.x, e.y, buffer.text))
                else
                    file:write(String.format("    { %d, %f, %f },", e.actorType, e.x, e.y))
                end
            end
        end)
        file:write("}")

        file:close()
    end

    function master.reset(self)
        -- Reset state
        self.timer = 0
        self.hours = 0
        self.minutes = 0
        self.work = 0

        -- Clear all children
        self:clearChildren()
        self.bosses = { }
        self:addBoard()

        -- Background
        local background = Entity.new(Master.boardWidth / 2, Master.boardHeight / 2, -2, Master.boardWidth, Master.boardHeight)

        background.elements:add(Element.Image.new("Images/Board.png"))
        self:addChild(background)

        -- Place actors
        if isTable(actors) then
            local i = 1

            while i <= #actors do
                local entry = actors[i]
                local actor = self:addChild(actorConstructors[entry[1]](self, entry))
                local actorType = actor.actorType

                if actorType == Actor.type.player then
                    self.player = actor
                elseif actorType == Actor.type.boss then
                    self.bosses[#self.bosses + 1] = actor
                end

                i = i + 1
            end
        end

        if debug then
            self.detected = false
        end
    end

    local size = 0.8
    local halfSize = size / 2

    function master.getValidX(self, e, x)
        -- Check overall bounds
        local x1, x2 = x - halfSize, x + halfSize

        if x1 < 0 then
            x = 0 + halfSize
        elseif x2 > Master.boardWidth then
            x = Master.boardWidth - halfSize
        end

        -- Check perpendicular walls
        local y1, y2 = e.y - halfSize, e.y + halfSize
        local i1, i2 = ceil(y1 + 0.001), ceil(y2 - 0.001)
        local i = i1

        while i <= i2 do
            if x > e.x then
                local tmp1 = x

                if wallsV[i][ceil(x) + 1] ~= nil then
                    x = min(x, ceil(x) - halfSize)
                end
            elseif x < e.x then
                if wallsV[i][ceil(x)] ~= nil then
                    x = max(x, floor(x) + halfSize)
                end
            end

            i = i + 1
        end

        -- Check parallel walls
        if i2 ~= i1 then
            if x > e.x and ceil(x) + 1 <= Master.boardWidth and wallsH[i1 + 1][ceil(x) + 1] ~= nil then
                x = min(x, ceil(x) - halfSize)
            elseif x < e.x and floor(x) >= 1 and wallsH[i1 + 1][floor(x)] ~= nil then
                x = max(x, floor(x) + halfSize)
            end
        end

        return x
    end

    function master.getValidY(self, e, y)
        -- Check overall bounds
        local y1, y2 = y - halfSize, y + halfSize

        if y1 < 0 then
            y = 0 + halfSize
        elseif y2 > Master.boardHeight then
            y = Master.boardHeight - halfSize
        end

        -- Check perpendicular walls
        local x1, x2 = e.x - halfSize, e.x + halfSize
        local j1, j2 = ceil(x1 + 0.001), ceil(x2 - 0.001)
        local j = j1

        while j <= j2 do
            if y > e.y then
                if wallsH[ceil(y) + 1][j] ~= nil then
                    y = min(y, ceil(y) - halfSize)
                end
            elseif y < e.y then
                if wallsH[ceil(y)][j] ~= nil then
                    y = max(y, floor(y) + halfSize)
                end
            end

            j = j + 1
        end

        -- Check parallel walls
        if j2 ~= j1 then
            if y > e.y and ceil(y) + 1 <= Master.boardHeight and wallsV[ceil(y) + 1][j1 + 1] ~= nil then
                y = min(y, ceil(y) - halfSize)
            elseif y < e.y and floor(y) >= 1 and wallsV[floor(y)][j1 + 1] ~= nil then
                y = max(y, floor(y) + halfSize)
            end
        end

        return y
    end

    local function forEachWall(f, ax, ay, bx, by)
        if bx ~= ax then
            local slope = (by - ay) / (bx - ax)
            local j1 = ceil(ax)
            local j2 = floor(bx)
            
            if bx < ax then
                j1 = ceil(bx)
                j2 = floor(ax)
            end
            
            local j = j1
            
            while j <= j2 do
                local i = floor(ay  + slope * (j - ax)) + 1
            
                if i < 1 or i > Master.boardHeight then
                    break
                end

                -- Supply the callback with the (potential) wall's position
                f(i, j)
            
                j = j + 1
            end
        end
    end

    function master.addWork(self)
        self.work = self.work + 1
    end

    function master.update(self, ms)
        if self.running then
            self:updateChildren(ms)

            -- Update time
            local done = false

            self.timer = self.timer + ms
            self.minutes = min(480, floor(self.timer / Master.minuteMS))
            done = self.minutes >= 480
            self.hours = floor(self.minutes / 60)
            self.minutes = self.minutes % 60

            -- Check for end
            if done then
                self.running = false

                if isFunction(self.completed) then
                    completed()
                end

                return
            end

            -- Check for detection
            local index = 1

            self.detected = false

            while index <= #self.bosses do
                local boss = self.bosses[index]

                if boss.active then
                    local f2 = Boss.awareFront / 2
                    local b2 = Boss.awareBack / 2
                    local d = angleToDirection(boss.lastAngle)
                    local k = 1

                    while k <= 2 do
                        local x1, x2, y1, y2

                        if k == 1 then
                            x1 = d[1] * f2 - f2
                            x2 = d[1] * f2 + f2
                            y1 = d[2] * f2 - f2
                            y2 = d[2] * f2 + f2
                        else
                            x1 = -d[1] * b2 - b2
                            x2 = -d[1] * b2 + b2
                            y1 = -d[2] * b2 - b2
                            y2 = -d[2] * b2 + b2
                        end

                        local debugElement = nil

                        if debug then
                            -- Show scanning range
                            if k == 1 then
                                boss.scanFront.x, boss.scanFront.y = x2 - f2, y2 - f2
                                debugElement = boss.scanFront
                            else
                                boss.scanBack.x, boss.scanBack.y = x2 - b2, y2 - b2
                                debugElement = boss.scanBack
                            end

                            debugElement.color.green = 0
                        end

                        -- Check for intersection with player
                        x1 = boss.x + x1
                        x2 = boss.x + x2
                        y1 = boss.y + y1
                        y2 = boss.y + y2

                        local playerX, playerY = self.player.x, self.player.y

                        if playerX >= x1 and playerX <= x2 and playerY >= y1 and playerY <= y2 then
                            -- The player is within the visible/aware range, so check for line of sight
                            -- Do this by determining which horizontal (and then vertical) walls intersect the line and
                            -- checking to see if there is, in fact, a wall
                            local clear = true

                            -- Vertical check
                            forEachWall(function(i, j)
                                local wall = wallsV[i][j + 1]

                                if wall ~= nil then
                                    clear = false

                                    if debug then
                                        wall.color = Color.new(0, 0, 1)
                                    end
                                end
                            end, boss.x, boss.y, playerX, playerY)

                            -- Horizontal check
                            forEachWall(function(j, i)
                                local wall = wallsH[i + 1][j]

                                if wall ~= nil then
                                    clear = false

                                    if debug then
                                        wall.color = Color.new(0, 0, 1)
                                    end
                                end
                            end, boss.y, boss.x, playerY, playerX)

                            -- Update color when debugging
                            if debug then
                                if clear then
                                    debugElement.color.green = 1
                                else
                                    debugElement.color.green = 0.5
                                end
                            end

                            if clear then
                                boss.task = Boss.tasks:force(boss, Boss.workTaskIndex)
                                self.detected = true
                            end
                        end
                        
                        k = k + 1
                    end
                end

                index = index + 1
            end
        end
    end

    return master
end

Editor = { }

function Editor.new()
    local editor = Entity.new(0, 0, 0.8)

    local function dToAngle(d)
        if d[1] == 1 and d[2] == 0 then
            return 0
        elseif d[1] == -1 and d[2] == 0 then
            return 180
        elseif d[1] == 0 and d[2] == 1 then
            return 90
        else
            return 270
        end
    end

    function editor.getXY(self)
        return self.j - 0.5, self.i - 0.5
    end

    function editor.getIJHorizontal(self)
        local i = self.i
        local j = self.j

        if self.d[1] > 0 then
            j = j + 1
        end

        if self.d[2] > 0 then
            i = i + 1
        end

        return i, j, self.d[1] == 0
    end

    editor.elements:add(Element.Image.new("Images/Blank.png", 0, 0, 0, 1, 1, 0, Color.new(0.8, 0.8, 0.8)))
    editor.elements:add(Element.Image.new("Images/Blank.png", 0.4, 0, 0, 0.2, 1, 0, Color.new(1, 1, 1)))
    editor.d = { 1, 0 }
    editor.angle = dToAngle(editor.d)
    editor.i, editor.j = 1, 1
    editor.x, editor.y = editor:getXY()

    function editor.snap(self)
        self.j = max(1, floor(self.j))
        self.i = max(1, floor(self.i))
        self.x, self.y = self:getXY()
    end

    function editor.move(self, dx, dy, state)
        if state.shift then
            dx = dx * 5
            dy = dy * 5
        end

        if state.fractional then
            self.j = max(1, min(Master.boardWidth, self.j + dx / 2))
            self.i = max(1, min(Master.boardHeight, self.i + dy / 2))
            self.x, self.y = self:getXY()
        else
            if self.d[1] == dx and self.d[2] == dy then
                self.j = max(1, min(Master.boardWidth, self.j + dx))
                self.i = max(1, min(Master.boardHeight, self.i + dy))
                self.x, self.y = self:getXY()
            else
                self.d = { dx, dy }
                self.angle = dToAngle(self.d)
            end
        end
    end

    function editor.toggle(self, master)
        master:toggleWall(self:getIJHorizontal())
    end

    return editor
end

Label = { }

function Label.new(x, y, lines)
    local label = Entity.new(x, y, 5, Grind.textWidth, Grind.textHeight)

    if isString(lines) then
        label.element = label.elements:add(Element.Text.new(lines))
    elseif isTable(lines) then
        local i = 1

        while i <= #lines do
            label.element = label.elements:add(Element.Text.new(lines[i], 0, -(i - 1)))
            i = i + 1
        end
    end

    return label
end

ValueLabel = { }

function ValueLabel.new(name, x, y, table, key, format, font, textWidth, textHeight)
    local valueLabel = Entity.new(x, y, 5, Grind.textWidth, Grind.textHeight)
    local template = ""

    if isString(name) then
        template = name .. ": "
    end

    if isNumber(textWidth) and isNumber(textHeight) then
        valueLabel.width, valueLabel.height = textWidth, textHeight
    end

    valueLabel.label = valueLabel.elements:add(Element.Text.new(template))
    valueLabel.lastValue = nil

    if isString(font) then
        valueLabel.label.font = font
    end

    function valueLabel.update(self, ms)
        local value = table[key]

        if value ~= self.lastValue then
            if isFunction(format) then
                self.label.text = format(value)
            else
                if value == nil then
                    self.label.text = template .. "nil"
                elseif isBoolean(value) then
                    if value then
                        self.label.text = template .. "true"
                    else
                        self.label.text = template .. "false"
                    end
                elseif isNumber(value) or isString(value) then
                    self.label.text = template .. value
                else
                    self.label.text = template .. "?"
                end
            end
        end
    end

    return valueLabel
end

GameLayer = { }

function GameLayer.new()
    local rate = 33

    if edit then
        rate = -1
    end

    local layer = Layer.new(rate)
    local background = layer:addChild(Entity.new(0, 0, -2, 640, 480))

    background.elements:add(Element.Image.new("Images/Background.png"))

    layer.master = layer:addChild(Master.new(layer, -288, -208, 384, 384, function() layer:stop() end))
    Grind.input:setupEventHandlers(layer)

    -- Clock
    layer:addChild(ValueLabel.new(nil, 128, 144, layer.master, "hours", function(t) return String.format("%02d", (t + Master.hoursOffset - 1) % 12 + 1) end, "Fonts/LED.png", 32, 64))
    layer:addChild(ValueLabel.new(nil, 192, 144, layer.master, "minutes", function(t) return String.format(":%02d", t) end, "Fonts/LED.png", 32, 64))

    if debug then
        layer:addChild(ValueLabel.new("Detected", 144, 48, layer.master, "detected"))
    end

    if edit then
        local state = {
            shift = false,
            fractional = false,
        }

        local activateHandlers = {
            { "Toggle Wall", nil, function(self, pressed) if pressed then self.master.editor:toggle(self.master) end end },
            { "Add Player", nil, function(self, pressed) if pressed then self.master:addChild(Player.new(self.master, self.master.editor:getXY())) end end },
            { "Add Boss", function(self)
                    state.fractional = true
                end, function(self, pressed)
                    if pressed then
                        if state.boss == nil then
                            state.boss = self.master:addChild(Boss.new(self.master, self.master.editor:getXY()))
                        else
                            local x, y = self.master.editor:getXY()

                            state.boss:addLandmark(x, y)

                            local e = Entity.new(x, y, 0.1, 0.5, 0.5)

                            e.elements:add(Element.Image.new("Images/Blank.png"))
                            e.color = Color.new(1, 0, 0)
                            self.master:addChild(e)
                        end
                    end
                end, function(self)
                    state.boss = nil
                    state.fractional = false
                    self.master.editor:snap()
                end },
        }

        local function switch()
            if not isNumber(layer.editorIndex) then
                layer.editorIndex = 0
            else
                -- Cleanup for previous handler
                local cleanup = activateHandlers[layer.editorIndex][4]

                if isFunction(cleanup) then
                    cleanup(layer)
                end
            end

            layer.editorIndex = (layer.editorIndex % #activateHandlers) + 1

            local entry = activateHandlers[layer.editorIndex]

            layer.editorLabel.text = entry[1]

            if isFunction(entry[2]) then
                entry[2](self)
            end

            layer.eventHandlers.activate = entry[3]
        end


        local label = Entity.new(200, 0, 1, 8, 12)

        layer.master.running = false
        layer.master.editor = Editor.new()
        layer.editorLabel = label.elements:add(Element.Text.new())
        layer:addChild(label)
    
        layer.eventHandlers = {
            up = function(self, pressed) if pressed then self.master.editor:move(0, 1, state) end end,
            down = function(self, pressed) if pressed then self.master.editor:move(0, -1, state) end end,
            left = function(self, pressed) if pressed then self.master.editor:move(-1, 0, state) end end,
            right = function(self, pressed) if pressed then self.master.editor:move(1, 0, state) end end,
            switch = function(self, pressed) if pressed then switch() end end,
            shift = function(self, pressed) state.shift = pressed end,
            cancel = function(self, pressed)
                if pressed then
                    -- Switch to ensure state is saved
                    switch()

                    -- Write to a file
                    self.master:save("Config.lua")

                    -- Exit
                    Layers.pop()
                end end,
        }

        switch()
    else
        layer.eventHandlersStart = {
            activate = function(self, pressed) if not pressed then self:start() end end,
            cancel = function(self, pressed) Layers.pop() end,
        }

        layer.eventHandlersStopped = {
            activate = function(self, pressed) if not pressed then self:reset() end end,
            cancel = function(self, pressed) Layers.pop() end,
        }

        layer.eventHandlersRunning = {
            up = function(self, pressed) self.master.player:setControlPressed("up", pressed) end,
            down = function(self, pressed) self.master.player:setControlPressed("down", pressed) end,
            left = function(self, pressed) self.master.player:setControlPressed("left", pressed) end,
            right = function(self, pressed) self.master.player:setControlPressed("right", pressed) end,
            cancel = function(self, pressed) Layers.pop() end,
        }
    end

    layer.prompt = Label.new(144, 48, {
--       123456789012345678901
        "Avoid the suits; they",
        "want to give you work",
        "",
        "Try to make it",
        "through the whole day",
        "with as few to-do's",
        "as possible",
        "",
        "Controls:",
        "-Arrow keys:  Move",
        "-Escape:      Quit",
        "-Space/Enter: Start",
    })

    layer.gameOver = Label.new(144, 48 - Grind.textHeight * 4, "Game Over")
    layer.gameOver.width = Grind.textWidth * 2
    layer.gameOver.height = Grind.textHeight * 2
    
    function layer.inputReceived(self, event, pressed)
        local eventHandler = layer.eventHandlers[event]
        local handled = false
    
        if isFunction(eventHandler) then
            eventHandler(self, pressed)
            handled = true
        end
    
        return handled
    end

    function layer.reset(self)
        self.master.running = false
        self.master:reset()
        self.eventHandlers = self.eventHandlersStart
        self:addChild(self.prompt)

        if self.scoreLabel ~= nil then
            self:removeChild(self.scoreLabel)
            self.scoreLabel = nil
        end

        self:removeChild(self.gameOver)
    end

    function layer.start(self)
        self:removeChild(self.prompt)
        self.master.running = true
        self.eventHandlers = self.eventHandlersRunning

        -- Add score display
        self.scoreLabel = layer:addChild(ValueLabel.new("To-do's", 144, 48, layer.master, "work"))
    end

    function layer.stop(self)
        self:addChild(self.gameOver)
        self.eventHandlers = self.eventHandlersStopped
    end

    function layer.shown(self)
        if edit then
            self.master:reset()
            self.master:addChild(self.master.editor)
        else
            self:reset()
        end
    end

    return layer
end

Layers.push(GameLayer.new())


